# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import pytest
import six

from pyout.field import Field
from pyout.field import Nothing
from pyout.field import StyleProcessors
from pyout.field import Truncater


def test_field_base():
    assert Field()("ok") == "ok        "
    assert Field(width=5, align="right")("ok") == "   ok"


def test_field_update():
    field = Field()
    field.width = 2
    assert field("ok") == "ok"


def test_field_processors():
    def pre(_, result):
        return result.upper()

    def post1(_, result):
        return "AAA" + result

    def post2(_, result):
        return result + "ZZZ"

    field = Field(width=6, align="center",
                  default_keys=["some_key", "another_key"])
    field.add("pre", "some_key", pre)
    field.add("post", "another_key", *[post1, post2])
    assert field("ok") == "AAA  OK  ZZZ"

    with pytest.raises(ValueError):
        field.add("not pre or post", "k")

    with pytest.raises(ValueError):
        field.add("pre", "not registered key")


@pytest.mark.parametrize("text",
                         ["", "-", "…"],
                         ids=["text=''", "text='-'", "text='…'"])
def test_something_about_nothing(text):
    nada = Nothing(text=text)
    assert not nada

    assert six.text_type(nada) == text
    assert "{:5}".format(nada) == "{:5}".format(text)
    assert "x" + nada == "x" + text
    assert nada + "x" == text + "x"


def test_truncate_mark_true():
    fn = Truncater(7, marker=True).truncate

    assert fn(None, "abc") == "abc"
    assert fn(None, "abcdefg") == "abcdefg"
    assert fn(None, "abcdefgh") == "abcd..."


def test_truncate_mark_string():
    fn = Truncater(7, marker="…").truncate

    assert fn(None, "abc") == "abc"
    assert fn(None, "abcdefg") == "abcdefg"
    assert fn(None, "abcdefgh") == "abcdef…"


def test_truncate_mark_short():
    fn = Truncater(2, marker=True).truncate
    assert fn(None, "abc") == ".."


def test_truncate_nomark():
    fn = Truncater(7, marker=False).truncate

    assert fn(None, "abc") == "abc"
    assert fn(None, "abcdefg") == "abcdefg"
    assert fn(None, "abcdefgh") == "abcdefg"


def test_style_value_type():
    fn = StyleProcessors.value_type

    assert fn(True) == "simple"
    assert fn("red") == "simple"
    assert fn({"lookup": {"BAD": "red"}}) == "lookup"

    interval = {"interval": [(0, 50, "red"), (50, 80, "yellow")]}
    assert fn(interval) == "interval"

    with pytest.raises(ValueError):
        fn({"unknown": 1})


def test_style_processor_render():
    sp = StyleProcessors()
    with pytest.raises(NotImplementedError):
        sp.render("key", "value")
